------------------------------------------------------------------------------
-- ESFA example: RandOps
-- John O'Donnell, 2013

-- Test the ESFA system by generating random  operations and checking
-- that invariants are preserved
------------------------------------------------------------------------------

module Main where

import Control.Monad.IO.Class
import Random
import Data.Array.ESFA.SimArch

------------------------------------------------------------------------------

-- Parameters to control the run (could use command line options instead)
circsize, nops :: Int
circsize = 14    -- number of cells = 2^circsize
nops = 10000      -- number of operations to perform

main :: IO ()
main = run circsize nops

run circsize nops =
  simulateESFM circsize initStatistics (script nops)

script :: Int -> StESFA Statistics Int ()
script nops =
  do liftIO (putStrLn "ESFA Example: RandOps")
     esfmInitialize
     esfmSetOption EsfmTraceOperation
     runops nops 1
     liftIO $ putStrLn "\nFinal state at end of script"
     liftIO $ putStrLn "Final state of machine"
     showState
     liftIO $ putStrLn "\nFinal statistics of the run"
     showStats


runops :: Int -> Int -> StESFA Statistics Int ()
runops limit i
  | i > limit = return ()
  | otherwise =
      do randUpdate i
         runops limit (i+1)

------------------------------------------------------------------------------

data Statistics =
  Statistics
    { updateStats :: UpdateStatistics
    } deriving Show

initStatistics =
  Statistics
    { updateStats = initUpdateStatistics
    }

data UpdateStatistics =
  UpdateStatistics
    { n_updates :: Int,
      n_updates_ok, n_updates_err :: Int,
      n_updates_Nothing :: Int
    } deriving Show

initUpdateStatistics :: UpdateStatistics
initUpdateStatistics =
  UpdateStatistics
    { n_updates = 0,
      n_updates_ok = 0,
      n_updates_err = 0,
      n_updates_Nothing = 0
    }

printUpdateStatistics =
  do stats <- getUserState
     let s = updateStats stats
     liftIO $ putStrLn "Update statistics"
     liftIO $ putStrLn ("  updates:         "  ++ show (n_updates s))
     liftIO $ putStrLn ("  correct updates: " ++ show (n_updates_ok s))
     liftIO $ putStrLn ("  system errors:   " ++ show (n_updates_err s))
     liftIO $ putStrLn ("  Nothing results: " ++ show (n_updates_Nothing s))

------------------------------------------------------------------------------

randUpdate opcount =
  do xs <- esfmGetSpecArrays
     let n = length xs :: Int
--     achoice <- liftIO $ randomRIO (0 :: Int, n-1)
     achoice <- liftIO $ randomRIO (0 :: Int, n)  -- allow some invalid anames
     ichoice <- liftIO $ randomRIO (10 :: Int, 19)
     vchoice <- liftIO $ randomRIO (100 :: Int, 1000)
--     liftIO (putStrLn ("randUpdate:  narrays=" ++ show n
--                         ++ "  achoice=" ++ show achoice
--                         ++ "  ichoice=" ++ show ichoice
--                         ++ "  vchoice=" ++ show vchoice
--                         ++ "\n"))
     randNonexistent <- liftIO $ randomRIO (n, 1000000)  -- name of non-array
     let oldarray =
          if achoice < n
            then fst (xs!!achoice)
            else randNonexistent
     a <- esfmUpdate ("a" ++ show opcount) oldarray (ichoice,vchoice)
     ok <- checkSpecMachine
     stats <- getUserState
     let s = updateStats stats
     let s' = s { n_updates = n_updates s + 1,
                  n_updates_ok =  if ok then n_updates_ok s + 1
                                        else n_updates_ok s,
                  n_updates_err = if ok then n_updates_err s
                                        else n_updates_err s + 1,
                  n_updates_Nothing =
                                  case a of
                                    Nothing -> n_updates_Nothing s + 1
                                    Just _ -> n_updates_Nothing s
                }              
     putUserState (stats {updateStats = s' })
     return ()

showStats =
  do s <- getUserState
     liftIO $ putStrLn (show s)
     liftIO $ putStrLn "\nStatistics"
     printUpdateStatistics

checkSpecMachine :: StESFA Statistics Int Bool
checkSpecMachine =
  do sas <- esfmGetSpecArrays
     mas <- esfmGetMachineArrays
     let n = length sas :: Int
     let ok = sas == mas
--     liftIO $ putStrLn ("checkSpecMachine, narrays = " ++ show n
--                        ++ show ok)
     if ok
       then do -- liftIO $ putStrLn "ok"
               return ok
       else do esfmPrintSpecArrays
               esfmPrintMachineArrays
               liftIO $ putStrLn "\n\n\nsas\n"
               liftIO $ putStrLn (show sas)
               liftIO $ putStrLn "\n\n\nmas\n"
               liftIO $ putStrLn (show mas)
               esfmPrintState
               liftIO $ putStrLn (take 72 (repeat '-'))
               return ok

showState =
  do esfmPrintState
     esfmPrintMachineArrays
     esfmPrintSpecArrays
     stats <- esfmAnalyzeMachineState
     liftIO (putStrLn (show stats))

{-
JUNK
{-
     randUpdate 100
     randUpdate 101
     randUpdate 102
     randUpdate 103
     randUpdate 104
     randUpdate 105
     randUpdate 106
-}

     Just a1 <- esfmUpdate "a1" esfmEmpty (500,501)
     checkSpecMachine
     Just a2 <- esfmUpdate "a2" a1  (600,601)
     checkSpecMachine
     Just a3 <- esfmUpdate "a3" a2  (500,555)
     checkSpecMachine

     xs <- esfmGetSpecArrays
     liftIO (putStrLn ("number of arrays = " ++ show (length xs)))

     r1 <- liftIO $ randomRIO ((0,8) :: (Int,Int))
     liftIO $ putStrLn (show r1)
     r1 <- liftIO $ randomRIO ((0,8) :: (Int,Int))
     liftIO $ putStrLn (show r1)
     r1 <- liftIO $ randomRIO ((0,8) :: (Int,Int))
     liftIO $ putStrLn (show r1)
     r1 <- liftIO $ randomRIO ((0,8) :: (Int,Int))
     liftIO $ putStrLn (show r1)
     r1 <- liftIO $ randomRIO ((0,8) :: (Int,Int))
     liftIO $ putStrLn (show r1)
     r1 <- liftIO $ randomRIO ((0,8) :: (Int,Int))
     liftIO $ putStrLn (show r1)
     r1 <- liftIO $ randomRIO ((0,8) :: (Int,Int))
     liftIO $ putStrLn (show r1)
     r1 <- liftIO $ randomRIO ((0,8) :: (Int,Int))
     liftIO $ putStrLn (show r1)
     r1 <- liftIO $ randomRIO ((0,8) :: (Int,Int))
     liftIO $ putStrLn (show r1)
     r1 <- liftIO $ randomRIO ((0,2) :: (Int,Int))
     liftIO $ putStrLn (show r1)
     r1 <- liftIO $ randomRIO ((0,20) :: (Int,Int))
     liftIO $ putStrLn (show r1)

     r1 <- liftIO $ randomRIO ((0,100) :: (Int,Int))
     liftIO $ putStrLn (show r1)

     r1 <- liftIO $ randomRIO ((0,0) :: (Int,Int))
     liftIO $ putStrLn (show r1)

     Just a1 <- esfmUpdate "a1" esfmEmpty (100, 101)
-- a1 = 100-101
     Just a2 <- esfmUpdate "a2" a1 (200, 201)
-- a2 = 200-201 100-101 
     Just a3 <- esfmUpdate "a3" a2 (300, 301)
-- a3 = 300-301 200-201 100-101 
     Just a4 <- esfmUpdate "a4" a2 (400, 401)
-- a4 = 400-401 200-201 100-101 
     Just a5 <- esfmUpdate "a5" a2 (500, 501)
-- a5 = 500-501 200-201 100-101 
     Just a6 <- esfmUpdate "a6" a3 (600, 601)
-- a6 = 600-601 300-301 200-201 100-101 
     Just a7 <- esfmUpdate "a7" a5 (700, 701)
-- a7 = 700-701 500-501 200-201 100-101 
     Just a8 <- esfmUpdate "a8" a1 (800, 801)
-- a8 = 800-801 100-101
     Just a9 <- esfmUpdate "a9" esfmEmpty (900, 901)
-- a9 = 900-901
-}
