
------------------------------------------------------------------------
-- DP1: Data parallel Version 1

{- The data parallel implementation will be presented in a sequence of
versions.  DP1, the first version, is the simplest: it demonstrates
the key ideas but glosses over some important details that are
addressed in later versions.

Treat state as a list of cells.  Don't worry about where in the memory
a word will be placed; just cons a new word onto list, or insert it
into list.  Thus, in this model, the size of the memory is unbounded
and we don't worry about exactly how memory is allocated. -}

-- DP1 representation

type StateDP1 = [CellDP1]


{- A cell corresponds to one memory location; it holds a complete
array element. -}

data CellDP1 = CellDP1
  { low, high :: ACode,
    ind       :: EltIndex,
    val       :: EltValue,
    mapName   :: AName,
    mapCode   :: ACode }
  deriving Show


initCellDP1 :: CellDP1
initCellDP1 = CellDP1
  { low      = 0,
    high     = 0,
    ind      = 0,
    val      = 0,
    mapName  = 0,
    mapCode  = 0 }


--  DP1 algorithm

encode :: StateDP1 -> AName -> ACode
encode [] 0 = 0
encode cs n =
  if n==0
    then 0
    else selectUnique [mapCode c | c <- cs, mapName c == n]



dp1update :: StateDP1 -> AName -> (EltIndex,EltValue) -> (StateDP1, AName)

dp1update cs aname (aidx,aval) = (cs', aname')
  where
    acode = encode cs aname
    acode' = acode + 1
    aname' = newName cs
    cs' = newc : map adjust cs
    adjust c =
        c { low = if low c > acode
                      then low c + 1
                      else low c,
            high = if high c >= acode
                      then high c + 1
                      else high c,
            mapCode = if mapCode c > acode
                         then mapCode c + 1
                         else mapCode c }
    newc = initCellDP1
            { low = acode', high = acode',
              ind = aidx, val = aval,
              mapName = aname', mapCode = acode' }



dp1lookup :: StateDP1 -> AName -> EltIndex -> Maybe EltValue

dp1lookup cs aname aidx = r where
  acode = encode cs aname
  cs' = [c | c <- cs, low c <= acode && acode <= high c
                      && ind c == aidx]
  (x,v) = foldr f (1000000, bigval)
            [(high c - low c, val c) | c <- cs']
  f (x1,v1) (x2,v2) = if x1 < x2 then (x1,v1) else (x2,v2)
  r = if v== bigval then Nothing else Just v
  bigval = 9999999


--  THIS IS A TEMPORARY HACK ???

newName :: StateDP1 -> AName
newName cs =  (1 + maximum (0 : [mapName c | c <- cs]))


--  DP1 Infrastructure functions

Some general functions to operate on the state.

selectUnique :: [a] -> a
selectUnique [] = error "empty list"
selectUnique [x] = x
selectUnique (x:y:ys) = error "multiple matches"

--  Testing DP1

dp1init = []         -- initial memory state for dp1 implementation
dp1empty = 0   -- the empty array for dp1 implementation


--  DP1 printing and testing

showAmapBinding :: CellDP1 -> String
showAmapBinding c =
  show (mapName c) ++ "->" ++ show (mapCode c)

showElement :: CellDP1 -> String
showElement c =
  show (low c) ++ "~" ++ show (high c)
    ++ "[" ++ show (ind c) ++ "]=" ++ show (val c)

dumpStateDP1 :: StateDP1 -> String -> IO ()
dumpStateDP1 cs lbl =
  do putStrLn lbl
     putStrLn "name->code map"
     putStrLn (concat [showAmapBinding c ++ "  " | c <- cs])
     putStrLn (concat [showElement c ++ "  " | c <- cs])


mytest1 =
  do putStrLn "mytest1"
     let s0 = []
     let empty = 0
     let (idx1,idx2) = (50, 54)

     putStrLn (take 30 (repeat '-'))
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     putStrLn (take 20 (repeat '.'))
     putStrLn (show (memDump dp1lookup (idx1,idx2) s0 [empty]))

     putStrLn (take 30 (repeat '-'))
     let (s1,a1) = dp1update s0 empty (50, 101)
     putStrLn "a1 = dp1update empty (50,101)"
--     showDp1update (s1,a1) "a1 = dp1update empty (50,101)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1

     putStrLn (take 30 (repeat '-'))
     putStrLn "(dp1lookup s1 a1 51))"
     putStrLn (show (dp1lookup s1 a1 (51)))

     putStrLn (take 30 (repeat '-'))
     let (s2,a2) = dp1update s1 a1 (50, 201)
     putStrLn "a2 = dp1update a1 (50,201)"
--     showDp1update (s2,a2) "a2 = dp1update a1 (200,201)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2

     putStrLn (take 30 (repeat '-'))
     let (s3,a3) = dp1update s2 a1 (54, 301)
     putStrLn  "a3 = dp1update a1 (54,301)"
--     showDp1update (s3,a3) "a3 = dp1update a1 (300,301)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2
     printArrayValue dp1lookup (idx1,idx2) s3 a3

     putStrLn (take 30 (repeat '-'))
     let (s4,a4) = dp1update s3 a2 (53, 401)
     putStrLn  "a4 = dp1update a2 (53,401)"
--     showDp1update (s4,a4) "a4 = dp1update a2 (400,401)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2
     printArrayValue dp1lookup (idx1,idx2) s3 a3
     printArrayValue dp1lookup (idx1,idx2) s4 a4

     putStrLn (take 30 (repeat '-'))
     let (s5,a5) = dp1update s4 a2 (52, 501)
     putStrLn "a5 = dp1update a2 (52,501)"
--     showDp1update (s5,a5) "a5 = dp1update a2 (52,501)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2
     printArrayValue dp1lookup (idx1,idx2) s3 a3
     printArrayValue dp1lookup (idx1,idx2) s4 a4
     printArrayValue dp1lookup (idx1,idx2) s5 a5

     putStrLn (take 30 (repeat '-'))
     let (s6,a6) = dp1update s5 a3 (54, 501)
     putStrLn "a6 = dp1update a3 (54,601)"
--     showDp1update (s6,a6) "a6 = dp1update a3 (54,601)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2
     printArrayValue dp1lookup (idx1,idx2) s3 a3
     printArrayValue dp1lookup (idx1,idx2) s4 a4
     printArrayValue dp1lookup (idx1,idx2) s5 a5
     printArrayValue dp1lookup (idx1,idx2) s6 a6

     putStrLn (take 30 (repeat '-'))
     putStrLn "(dp1lookup s1 a1 100))"
     putStrLn (show (dp1lookup s1 a1 (53)))

     putStrLn (take 30 (repeat '-'))
     putStrLn "(dp1lookup s1 a5 500))"
     putStrLn (show (dp1lookup s6 a5 (52)))

     putStrLn (take 30 (repeat '-'))
     printArrayValue dp1lookup (idx1,idx2) s6 a6



-- dp1 manual testing
-- A collection of tests, using some alternative tools.  Not -- for
-- building real test suites.

bigtest =
  do putStrLn "bigtest"
     let s0 = []
     let empty = 0
     let (idx1,idx2) = (50, 54)

     putStrLn (take 30 (repeat '-'))
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     putStrLn (take 20 (repeat '.'))
     putStrLn (show (memDump dp1lookup (idx1,idx2) s0 [empty]))

     putStrLn (take 30 (repeat '-'))
     let (s1,a1) = dp1update s0 empty (50, 101)
     putStrLn "a1 = dp1update empty (50,101)"
--     showDp1update (s1,a1) "a1 = dp1update empty (50,101)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1

     putStrLn (take 30 (repeat '-'))
     putStrLn "(dp1lookup s1 a1 51))"
     putStrLn (show (dp1lookup s1 a1 (51)))

     putStrLn (take 30 (repeat '-'))
     let (s2,a2) = dp1update s1 a1 (50, 201)
     putStrLn "a2 = dp1update a1 (50,201)"
--     showDp1update (s2,a2) "a2 = dp1update a1 (200,201)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2

     putStrLn (take 30 (repeat '-'))
     let (s3,a3) = dp1update s2 a1 (54, 301)
     putStrLn  "a3 = dp1update a1 (54,301)"
--     showDp1update (s3,a3) "a3 = dp1update a1 (300,301)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2
     printArrayValue dp1lookup (idx1,idx2) s3 a3

     putStrLn (take 30 (repeat '-'))
     let (s4,a4) = dp1update s3 a2 (53, 401)
     putStrLn  "a4 = dp1update a2 (53,401)"
--     showDp1update (s4,a4) "a4 = dp1update a2 (400,401)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2
     printArrayValue dp1lookup (idx1,idx2) s3 a3
     printArrayValue dp1lookup (idx1,idx2) s4 a4

     putStrLn (take 30 (repeat '-'))
     let (s5,a5) = dp1update s4 a2 (52, 501)
     putStrLn "a5 = dp1update a2 (52,501)"
--     showDp1update (s5,a5) "a5 = dp1update a2 (52,501)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2
     printArrayValue dp1lookup (idx1,idx2) s3 a3
     printArrayValue dp1lookup (idx1,idx2) s4 a4
     printArrayValue dp1lookup (idx1,idx2) s5 a5

     putStrLn (take 30 (repeat '-'))
     let (s6,a6) = dp1update s5 a3 (54, 501)
     putStrLn "a6 = dp1update a3 (54,601)"
--     showDp1update (s6,a6) "a6 = dp1update a3 (54,601)"
     printArrayValue dp1lookup (idx1,idx2) s0 empty
     printArrayValue dp1lookup (idx1,idx2) s1 a1
     printArrayValue dp1lookup (idx1,idx2) s2 a2
     printArrayValue dp1lookup (idx1,idx2) s3 a3
     printArrayValue dp1lookup (idx1,idx2) s4 a4
     printArrayValue dp1lookup (idx1,idx2) s5 a5
     printArrayValue dp1lookup (idx1,idx2) s6 a6

     putStrLn (take 30 (repeat '-'))
     putStrLn "(dp1lookup s1 a1 100))"
     putStrLn (show (dp1lookup s1 a1 (53)))

     putStrLn (take 30 (repeat '-'))
     putStrLn "(dp1lookup s1 a5 500))"
     putStrLn (show (dp1lookup s6 a5 (52)))

     putStrLn (take 30 (repeat '-'))
     printArrayValue dp1lookup (idx1,idx2) s6 a6

------------------------------------------------------------------------
-- This test prints out internal state, and isn't suitable for user
--testing

run1 = run dp1update dp1lookup testscript1

testscript1 :: [Operation]
testscript1 =
  [Update (Avarnum 0) (50) (500),
   Update (Avarnum 0) (51) (600),
   Update (Avarnum 1) (52) (700),
   Lookup (Avarnum 1) (50),
   Lookup (Avarnum 2) (51),
   Lookup (Avarnum 3) (50),
   Lookup (Avarnum 3) (51),
   Lookup (Avarnum 3) (52)]

------------------------------------------------------------------------
-- Some helper definitions, user can ignore these

upd mst a i v = dp1update mst a (i, v)
s0 = dp1init
empty = dp1empty

------------------------------------------------------------------------


------------------------------------------------------------------------
-- DPLST: List model of circuit parallel implementation

{- Maybe this section should be deleted; the tree machine version is
realistic, unlike this one. -}

\begin{code}
type CpMachineState = [CpCell]               -- state of esf memory
type CpSt a = StateT CpMachineState IO a     -- circuit state monad
\end{code}

\subsection{CP machine structure}
\label{sec:cp-machine-structure}

\begin{code}
data CpCell = CpCell
  { cpFull         :: Bool,
    cpLow, cpHigh  :: ACode,
    cpInd          :: EltIndex,
    cpVal          :: EltValue,
    cpMapName      :: AName,
    cpMapCode      :: ACode }
  deriving Show
\end{code}

\begin{code}
putInd :: CpCell -> EltIndex -> CpCell
putInd c x = c {cpInd = x}
\end{code}

\begin{code}
initCpCell :: CpCell
initCpCell = CpCell
  { cpFull     = False,
    cpLow      = 0,
    cpHigh     = 0,
    cpInd      = 0,
    cpVal      = 0,
    cpMapName  = 0,
    cpMapCode  = 0 }
\end{code}

\subsection{CP machine operations}
\label{sec:cp-machine-operations}

\begin{code}
cpmap :: Monad m => (a->c->c) -> a -> StateT [c] m ()
cpmap f x =
  do cells <- get
     put (map (f x) cells)
     return ()
\end{code}

\begin{code}
cpfold :: Monad m => (c->a) -> (a->a->a) -> StateT [c] m a
cpfold f g =
  do cells <- get
     return (foldl1 g (map f cells))
\end{code}

g needs to be associative

\begin{code}
cpscanl :: Monad m => (c->a) -> (a->a->a) -> (c->a->c) -> a -> StateT [c] m a
cpscanl f g h x =
  do cells <- get
     let xs = scanl g x (map f cells)
     let xs' = take (length cells) xs
     let [y] = drop (length cells) xs
     put (zipWith h cells xs')
     return y
\end{code}

\begin{code}
printState :: StateT [CpCell] IO ()
printState =
  do s <- get
     liftIO (putStrLn (show s))
     return ()
\end{code}

\subsection{CP implementation of array operations}
\label{sec:cp-array-operations}

\begin{code}
getArrayCode :: AName -> StateT [CpCell]  IO (Maybe ACode)
getArrayCode an =
  do ac <- cpfold (cellNameStatus an) foo
     return ac

cellNameStatus :: AName -> CpCell -> Maybe ACode
cellNameStatus an c =
  if an == cpMapName c
    then Just (cpMapCode c)
    else Nothing

foo Nothing Nothing = Nothing  -- surely a standard function?
foo (Just x) Nothing = Just x
foo Nothing (Just y) = Just y
foo (Just x) (Just y) = Just x
\end{code}

\begin{verbatim}
cpUpdate :: AName -> AIndex -> AValue -> StateT [CpCell] IO (Maybe AName)
cpUpdate an ai av =
  do 
\end{verbatim}

\subsection{Running the CP algorithm}
\label{sec:running-cp-algorithm}

For testing, a function that can be mapped

\begin{code}
silly :: Int -> CpCell -> CpCell
silly i x =
  let c = cpHigh x
  in x {cpHigh = c+i}
\end{code}

\begin{verbatim}
test_map :: StateT [CpCell] IO ()
test_map =
  do liftIO (putStrLn "test_map")
     liftIO (putStrLn "Here is the initial state")
     printState
     cpmap silly 3
     liftIO (putStrLn "Just did  cpmap silly 3")
     printState
     cpmap silly 9
     liftIO (putStrLn "Just did  cpmap silly 9")
     printState
     return ()

{-
test_scanl :: StateT [CpCell] IO ()
test_scanl =
  do liftIO (putStrLn "test_scanl")
     liftIO (putStrLn "Here is the initial state")
     printState
     cpmap silly 3
     liftIO (putStrLn "Just did  cpmap silly 3")
     printState
     cpscanl cpInd (+) putInd 100
     liftIO (putStrLn "Just did  cpscanl")
     printState
     return ()
-}

{-
addInd :: Int -> EltIndex -> I
-}

run_scanl :: IO ()
run_scanl = cpgo 5 test_scanl

cpgo :: Int -> StateT [CpCell] IO () -> IO ()
cpgo n f =
  do putStrLn "Running cpgo"
     let initialState = take n (repeat initCpCell)
     runStateT f initialState
     putStrLn "bye"
\end{verbatim}

