

{-# OPTIONS_GHC -XMultiParamTypeClasses #-}
{-# OPTIONS_GHC -XFunctionalDependencies #-}
{-# OPTIONS_GHC -XFlexibleInstances #-}

module Data.Array.ESFA.Spec where
import Control.Monad.IO.Class
import Control.Monad.Trans.State.Strict
import Data.List
import Data.Ord
import Test.QuickCheck

{- A generalisation of pure functional arrays is defined, which
  supports extensibility and sparsity.  A data parallel algorithm can
  implement the array operations efficiently. -}


-- Specification

-- Basic pure operations

{-
This section introduces a basic version: ignoring state, no
provision for deletion, no handling of errors.

Arrays have type a, with index of type i and element values of type
e.  A basic ESF array system has a constant empty array called
empty, and two functions for updating an array and lookup up a
value in an array.

Primitive types used in the representation.
-}

type EltIndex = Int
type EltValue = Int
type AName    = Int
type ACode    = Int


type Esfa = Int
--  represents an esf array, with Int indexes, and Word values



-- esfEmpty :: ESFA
--  The empty array, a constant that is always available.

-- esfUpdate :: ESFA -> i -> v -> ESFA
--  Update an array with an index i and value v; result is a new array

-- esfLookup :: ESFA -> i -> Maybe v

-- Lookup an array at index i; if defined the result is Just v; if
-- undefined the result is Nothing.

-- esfDelete :: ESFA -> ()

-- Semantically, this removes the correctness constraints on a lookup
-- to the array.  Operationally, it removes the array from the system
-- and reclaims its storage.

{-

Axiom 1. (Empty array)
   esfLookup esfEmpty i = Nothing

Axiom 2. (Nonempty array)
   esfLookup (esfUpdate a j v) i
     | i==j  = Just v
     | i/=j  = esfLookup a j

This specification is too simplistic to capture some essential
properties, including keeping the ESF arrays in a separate memory
and allowing for storage management, in particular deletion of
arrays.
-}

-- Operations on ESF arrays

{- General operations, allowing for a state containing the arrays and
allowing operations to fail, and providing an operation to delete an
array.

Again, a is the type of an ESF array; i is the index type, and v is
the element type.  In addition, the ESF arrays are held in a separate
memory with state of type s.  -}


class Eq i => EsfaM s a i v | a -> i, a -> v where
  esfEmpty :: a
  esfUpdate :: s -> a -> i -> v -> (s, Maybe a)
  esfLookup :: s -> a -> i -> Maybe v
  esfDelete :: s -> a -> s

{- Both lookup and update can fail, giving Nothing.  Lookup fails if
the array a is not defined, or if a is defined but does not have a
defined value at index i.  Update fails if the array a is not defined,
or if the memory does not have enough space to allocate a new
element. -}

-- Axioms

{- Improved axioms, including state, allowing deletion and allowing
update to fail (if the memory is full) -}

{-
Axiom 1: lookup on empty array
   esfLookup s esfEmpty i = Nothing

  A lookup on an empty array always gives Nothing, regardless of the
  state.

Axiom 2: lookup on nonempty array

  let (s2,a2) = esfUpdate s1 a j v
  r = esfLookup s2 a i
Then r =
  if  i==j  then Just v
  if  i/=j  then esfLookup a j

-}

-- Reference implementation using datatype: AxArr

{- An ESF array is either empty or an update of an existing ESF
array. -}

data AxArr i v
  = AxEmpty
  | AxUpdate (AxArr i v) i v

{- The axLookup and axUpdate functions are defined directly according
to the axioms. -}

axUpdate :: s -> AxArr i v -> i -> v -> (s, Maybe (AxArr i v))
axUpdate s a i v = (s, Just (AxUpdate a i v))

axLookup :: Eq i => s -> AxArr i v -> i -> Maybe v
axLookup s AxEmpty i = Nothing
axLookup s (AxUpdate a j x) i
  | i==j = Just x
  | i/=j = axLookup s a i


instance Eq i =>  EsfaM s (AxArr i v) i v where
  esfEmpty = AxEmpty
  esfUpdate = axUpdate
  esfLookup = axLookup
  esfDelete s a = s

test_ax :: IO ()
test_ax =
  do putStrLn "\ntest_ax"
     let empty = esfEmpty :: AxArr Int Int
     let s0 = ()
     let (s1, Just a1) = esfUpdate s0 empty 30 35
     let (s2, Just a2) = esfUpdate s1 a1 40 50
     let fet35 = esfLookup s2 a2 30
     let fet50 = esfLookup s2 a2 40
     putStrLn (show fet35)
     putStrLn (show fet50)
--     check (esfLookup s2 a2 30) (Just 35)
--     check (esfLookup s2 a2 40) (Just 50)
--     check (esfLookup s2 a2 50) Nothing
     putStrLn "\nend test_ax"


-- History and deletions

{- The pure specification of update and lookup say nothing about
deleted arrays.  A delete operation is issued by the main program,
presumably by the garbage collector.  It should have the effect of
removing information belonging only to the deleted array from the ESF
memory.  It also has the effect of removing the obligation that lookup
a i gives a defined value after a has been deleted.

To model deletions, it's necessary to maintain the ESF memory state.
Also, an array returned should be a compact value (i.e. a word) rather
than a large data structure linked through the main heap.


After a sequence of operations, there is a set of arrays; each is
identified by an array reference.  An array reference is a compact
value, and can be recycled after the array it refers to is deleted. -}

------------------------------------------------------------------------
-- Macro operations

{- arrayValue gives the entire value of an array between two index
bounds; i.e. the set of bindings. -}

arrayValue
  :: (s -> a -> EltIndex -> Maybe EltValue)  -- lookup
  -> (EltIndex,EltIndex)                     -- index range
  -> s                                       -- state
  -> a                                       -- array
  -> [(EltIndex,Maybe EltValue)]             -- contents of array

arrayValue flookup (j, k) s aname =
  let f i
        | i > k = []
        | i <= k =
            (i, flookup s aname i) : f (i+1)
  in f j


memDump
  :: (s -> a -> EltIndex -> Maybe EltValue)   -- lookup
  -> (EltIndex,EltIndex)                      -- index range
  -> s                                        -- state
  -> [a]                                      -- all the arrays
  -> [[(EltIndex,Maybe EltValue)]]            -- contents of arrays

memDump flookup (j,k) s anames =
  map (arrayValue flookup (j,k) s) anames

-- Showing and printing arrays

showArrayValue :: Show a => a -> [(EltIndex, Maybe EltValue)] -> String
showArrayValue an xs =
  "Array " ++ show an ++ " contains\n" ++
    concat (map (("  "++) . showABinding) xs)

showABinding :: (EltIndex, Maybe EltValue) -> String
--showABinding (i,v) = "[" ++ show i ++ "]=" ++ show v
showABinding (i, Nothing) = show i ++ "->X"
showABinding (i, Just v) =
  show i ++ "->" ++ show v

printArrayValue
  :: (s -> AName -> EltIndex -> Maybe EltValue)
  -> (EltIndex,EltIndex)
  -> s -> AName
  -> IO ()
printArrayValue flookup (j,k) s aname =
  do let ys = arrayValue flookup (j,k) s aname
     putStrLn (showArrayValue aname ys)
     putStrLn []


------------------------------------------------------------------------
--  LST: List implementation

-- ESF array operations
  
{- This is a stateless specification, representing an array as a list
of updates.

Use i for index (type and value of index), v for element (type and
value of element).  An element binding is a pair (i,v) and an array
is a list of bindings [(i,v)].  There is an initial array empty =
[].

A list implementation with element type b and value type c -}


type EsfList i v = [(i,v)]


listEmpty :: [(i,v)]
listEmpty = []

listUpdate :: s -> [(i,v)] -> i -> v -> (s,[(i,v)])
listUpdate s a i x = (s, (i,x):a)

listLookup :: Eq i => s -> [(i,v)] -> i -> Maybe v
listLookup s [] i = Nothing
listLookup s ((j,x):a) i
  | j==i = Just x
  | otherwise = listLookup s a i


--  instance Eq b => ESFA () [(b,c)] b c where

{-
instance Eq i => ESFA s [(i,v)] i v where
  esfEmpty  = listEmpty
  esfUpdate = listUpdate
  esfLookup = listLookup
-}

{-
test_list :: IO ()
test_list =
  do putStrLn "\ntest_list implementation"
     let empty = esfEmpty :: [(Int,Int)]
     let s0 = ()
     let (s1,a1) = esfUpdate s0 empty 30 35
     let (s2,a2) = esfUpdate s1 a1 40 50
     let fet35 = esfLookup s2 a2 30
     let fet50 = esfLookup s2 a2 40
     putStrLn (show fet35)
     putStrLn (show fet50)
     check (esfLookup s2 a2 30) (Just 35)
     check (esfLookup s2 a2 40) (Just 50)
     check (esfLookup s2 a2 50) Nothing
     putStrLn "\nend test test list implementation"
-}

{- The implementations that will be developed represent the arrays in
a state, and the operations perform updates an lookups on the state.
Furthermore, there are equivalence properties that should hold between
the specification and the implementation.  Therefore it is useful to
have an equivalent form of the specification of the operations using a
dummy state, so all the definitions have consistent types. -}

listUpdatewithSt s a (i,x) = (s, (i,x) : a)

listLookupwithSt s [] i = Nothing
listLookupwithSt s ((j,x):a) i
  | j==i = Just x
  | otherwise = listLookupwithSt s a i


